@@ -377,7 +377,7 @@ def make_new_vis_cube(res, n_pix, freqs, umin, umax, kern=None, K=None, uv_bins_
     uu, vv, _ = psutil.get_ungrid_vis_idx((n_pix, n_pix), res, umin, umax)
 
     meta = datacube.ImageMetaData.from_res(res, (n_pix, n_pix))
-    meta.wcs.wcs.cdelt[2] = np.diff(freqs).mean()
+    meta.wcs.wcs.cdelt[2] = psutil.robust_freq_width(freqs)
 
     c = datacube.CartDataCube(np.zeros((len(freqs), len(uu)), dtype=np.complex128), uu, vv, freqs, meta)
 
@@ -672,11 +672,18 @@ class AbstractMLKern(MultiKern, GPy.kern.Kern):
         norm_cov_1d = self.get_norm_cov_1d(freqs, np.array(self.params[:-1]).T)
         r_1d = dnu_mhz * np.arange(norm_cov_1d.shape[1])
 
-        Xsq = np.sum(np.square(X), 1)
-        r2 = -2. * np.dot(X, X.T) + (Xsq[:, None] + Xsq[None, :])
-        r2[np.diag_indices(X.shape[0])] = 0.
-        r2 = np.clip(r2, 0, np.inf)
-        r = np.sqrt(r2)
+        if X2 is None:
+            Xsq = np.sum(np.square(X), 1)
+            r2 = -2. * np.dot(X, X.T) + (Xsq[:, None] + Xsq[None, :])
+            r2[np.diag_indices(X.shape[0])] = 0.
+            r2 = np.clip(r2, 0, np.inf)
+            r = np.sqrt(r2)
+        else:
+            Xsq = np.sum(np.square(X), 1)
+            X2sq = np.sum(np.square(X2), 1)
+            r2 = -2. * np.dot(X, X2.T) + (Xsq[:, None] + X2sq[None, :])
+            r2 = np.clip(r2, 0, np.inf)
+            r = np.sqrt(r2)
 
         cov = scipy.interpolate.interp1d(r_1d, norm_cov_1d, kind='quadratic', axis=1,
                                          bounds_error=False, fill_value=0)(r)
@@ -872,23 +879,33 @@ class MExponential(MutliStationaryKern, MultiKern, GPy.kern.Exponential):
 
 class MWhiteHeteroscedastic(MultiKern, GPy.kern.Kern):
 
-    def __init__(self, variance=1, name='noise'):
+    def __init__(self, variance=1, X=None, name='noise'):
         MultiKern.__init__(self)
         GPy.kern.Kern.__init__(self, 1, 0, name)
-        self.set_variance(variance)
+        self.set_variance(variance, X=X)
         self.alpha = GPy.core.parameterization.Param('alpha', 1)
         self.alpha.constrain_fixed()
         self.link_parameters(self.alpha)
 
-    def set_variance(self, variance):
+    def set_variance(self, variance, X=None):
         assert np.isscalar(variance) or (len(self.uv_bins) == len(variance))
+        if not np.isscalar(variance):
+            assert X is not None, 'X positions must be informed with Heteroscedastic noise'
         self.variance = variance
+        self.X = X
 
     def K(self, X, X2=None):
+        if X2 is None:
+            X2 = X
         if np.isscalar(self.variance):
-            Ks = np.array([np.eye(X.shape[0]) * self.variance] * len(self.uv_bins))
+            variance = self.variance
         else:
-            Ks = np.array([np.diag(v) for v in self.variance])
+            x_interp = scipy.interpolate.interp1d(self.X[:, 0], self.variance, bounds_error=False, fill_value='extrapolate')
+            variance = x_interp(X[np.in1d(X, X2)][:, 0])
+
+        Ks = np.zeros((len(self.uv_bins), len(X), len(X2)))
+        Ks[:, np.in1d(X, X2), np.in1d(X2, X)] = variance
+
         return self.alpha[0] * Ks
 
 
@@ -966,8 +983,7 @@ class Exponent10(paramz.transformations.Transformation):
 class GPRegressor:
     ''' Simple GPR. See e.g. http://www.gaussianprocess.org/gpml/chapters/RW2.pdf'''
 
-    def __init__(self, X, Y, K):
-        self.X = X
+    def __init__(self, Y, K):
         self.K = K
         self.Y = Y
         self.fit()
@@ -976,10 +992,15 @@ class GPRegressor:
         self.L_ = GPy.util.linalg.jitchol(self.K, maxtries=100)
         self.alpha_, _ = GPy.util.linalg.dpotrs(self.L_, self.Y, lower=1)
 
-    def predict(self, K_star):
-        y_mean = K_star.dot(self.alpha_)
-        v, _ = GPy.util.linalg.dpotrs(self.L_, K_star.T, lower=1)
-        y_cov = K_star - K_star.dot(v)
+    def predict(self, K_p=None, K_p_p=None):
+        if K_p is None:
+            K_p = self.K
+            K_p_p = self.K
+        else:
+            assert K_p_p is not None, "K_p_p needs to be given when predicting at different X"
+        y_mean = K_p.T.dot(self.alpha_)
+        v, _ = GPy.util.linalg.dpotrs(self.L_, K_p, lower=1)
+        y_cov = K_p_p - K_p.T.dot(v)
         return y_mean, y_cov
 
     def log_marginal_likelihood(self):
@@ -1027,11 +1048,23 @@ class MultiData(object):
         else:
             return self.variance_split(self.noise_cube)
 
-    def gen_cube(self, Ys_and_covYs):
-        cube = self.i_cube.copy()
+    def get_freqs(self, fill_gaps=False):
+        freqs = self.i_cube.freqs
+        if fill_gaps:
+            freqs = np.array(sorted(np.concatenate((freqs, psutil.get_freqs_gaps(freqs)))))
+        return freqs
+
+    def gen_cube(self, Ys_and_covYs, fill_gaps=False):
+        freqs_p = self.get_freqs(fill_gaps=fill_gaps)
+        if fill_gaps:
+            freqs_p = np.array(sorted(np.concatenate((freqs_p, psutil.get_freqs_gaps(freqs_p)))))
+        X_p = (freqs_p * 1e-6)[:, None]
+        data = np.zeros((len(freqs_p), self.i_cube.data.shape[1]), dtype=self.i_cube.data.dtype)
+        cube = self.i_cube.new_with_data(data, freqs=freqs_p)
+
         idxs = [(cube.ru >= umin) & (cube.ru < umax) for umin, umax in self.uv_bins]
         for (Y, covY), idx in zip(Ys_and_covYs, idxs):
-            cube.data[:, idx] = self.f2c(Y) + get_samples(self.X, idx.sum(), covY)
+            cube.data[:, idx] = self.f2c(Y) + get_samples(X_p, idx.sum(), covY)
         return 1 / self.norm_factor * cube
 
 
@@ -1055,7 +1088,7 @@ class MultiGPRegressor(GPy.core.model.Model):
 
         kern_model.set_uv_bins(multi_data.uv_bins)
         kern_noise.set_uv_bins(multi_data.uv_bins)
-        kern_noise.set_variance(multi_data.noise_variance_split())
+        kern_noise.set_variance(multi_data.noise_variance_split(), multi_data.X)
 
         self.multi_data = multi_data
         self.kern_model = kern_model
@@ -1076,18 +1109,33 @@ class MultiGPRegressor(GPy.core.model.Model):
     def fit(self):
         Ks = self.kern.K(self.X)
         if Ks.ndim == 2 and self.Ys.ndim == 2:
-            self.gp_regressors = [GPRegressor(self.X, self.Ys, Ks)]
+            self.gp_regressors = [GPRegressor(self.Ys, Ks)]
         else:
-            self.gp_regressors = [GPRegressor(self.X, Y, K) for K, Y in zip(Ks, self.Ys)]
+            self.gp_regressors = [GPRegressor(Y, K) for K, Y in zip(Ks, self.Ys)]
 
-    def predict(self, kern=None):
+    def predict(self, kern=None, fill_gaps=False):
         assert self.gp_regressors is not None
         if kern is None:
             kern = self.kern_model
-        Ks = kern.K(self.X)
-        if Ks.ndim == 2:
-            Ks = np.repeat(Ks[None], len(self.Ys), axis=0)
-        return self.multi_data.gen_cube([gp_regressor.predict(K) for K, gp_regressor in zip(Ks, self.gp_regressors)])
+        if fill_gaps:
+            freqs_p = self.multi_data.get_freqs(fill_gaps=fill_gaps)
+            X_p = (freqs_p * 1e-6)[:, None]
+            Ks_p = kern.K(self.X, X_p)
+            Ks_p_p = kern.K(X_p, X_p)
+        else:
+            Ks_p = kern.K(self.X)
+            Ks_p_p = Ks_p
+        if Ks_p.ndim == 2:
+            Ks_p = np.repeat(Ks_p[None], len(self.Ys), axis=0)
+            Ks_p_p = np.repeat(Ks_p_p[None], len(self.Ys), axis=0)
+        return self.multi_data.gen_cube([gp_regressor.predict(K_p, K_p_p) for K_p, K_p_p, gp_regressor in zip(Ks_p, Ks_p_p, self.gp_regressors)],
+                                         fill_gaps=fill_gaps)
+
+    def get_interpolated_i_cube(self):
+        full_predicted_cube = self.predict(self.kern, fill_gaps=True)
+        idx = np.in1d(self.multi_data.get_freqs(True), self.multi_data.get_freqs(False))
+        full_predicted_cube.data[idx] = self.multi_data.i_cube.data
+        return full_predicted_cube
 
     def log_marginal_likelihood(self):
         assert self.gp_regressors is not None
@@ -1159,6 +1207,8 @@ class MCMCSampler(object):
 
         if emcee_moves == 'stretch':
             moves = emcee.moves.StretchMove()
+        elif emcee_moves == 'kde':
+            moves = emcee.moves.KDEMove()
         else:
             moves = emcee_moves
         if save_file is not None:
@@ -1320,29 +1370,42 @@ class MCMCSampler(object):
         return corner.corner(samples, plot_datapoints=False, smooth=0.8, 
                              quantiles=(0.16, 0.84), labels=self.get_parameter_names())
 
-    def get_ps_stack(self, ps_gen, kbins, n_burn, n_pick=100, kern_name='eor',
-                     subtract_from=None, verbose=False, clip_nsigma=6, discard_walkers_nsigma=10):
-        ps_stacker = pspec.PsStacker(ps_gen, kbins)
+    def select_random_sample(self, samples):
+        i = np.random.randint(0, samples.shape[0], 1)[0]
+        m_oa = samples[i].copy()
+        self.gp.kern.optimizer_array = m_oa
+        self.gp.fit()
+
+    def generate_data_cubes(self, n_burn, n_pick, kern_name='eor', clip_nsigma=6, discard_walkers_nsigma=10, fill_gaps=False):
         samples = self.get_samples(n_burn=n_burn, clip_nsigma=clip_nsigma, 
                                    discard_walkers_nsigma=discard_walkers_nsigma)
-        pick = np.random.randint(0, samples.shape[0], n_pick)
-        pr = psutil.progress_report(len(pick))
-        
+
         # Determine the parameters index of the portion of the kernel that will be predicted
         k = self.gp.kern.copy()
         k.optimizer_array = np.arange(len(k.optimizer_array))
         k_part = get_kern_part(k, kern_name)
         k_part_idx = k_part.optimizer_array.astype(int)
 
-        for j, i in enumerate(pick):
+        for i in range(n_pick):
+            self.select_random_sample(samples)
+            k_part.optimizer_array = self.gp.kern.optimizer_array[k_part_idx]
+            yield self.gp.predict(k_part, fill_gaps=fill_gaps)
+
+    def get_interpolated_i_cube(self, n_burn, clip_nsigma=6, discard_walkers_nsigma=10):
+        samples = self.get_samples(n_burn=n_burn, clip_nsigma=clip_nsigma,  discard_walkers_nsigma=discard_walkers_nsigma)
+        self.select_random_sample(samples)
+        return self.gp.get_interpolated_i_cube()
+
+    def get_ps_stack(self, ps_gen, kbins, n_burn, n_pick=100, kern_name='eor',
+                     subtract_from=None, clip_nsigma=6, discard_walkers_nsigma=10, fill_gaps=False):
+        ps_stacker = pspec.PsStacker(ps_gen, kbins)
+        pr = psutil.progress_report(n_pick)
+        
+        generator = self.generate_data_cubes(n_burn, n_pick, kern_name=kern_name, clip_nsigma=clip_nsigma,
+                                             discard_walkers_nsigma=discard_walkers_nsigma, fill_gaps=fill_gaps)
+
+        for j, c_rec in enumerate(generator):
             pr(j)
-            m_oa = samples[i].copy()
-            self.gp.kern.optimizer_array = m_oa
-            k_part.optimizer_array = m_oa[k_part_idx]
-            self.gp.fit()
-            if verbose:
-                print(m_oa)
-            c_rec = self.gp.predict(k_part)
             if subtract_from is not None:
                 c_rec = subtract_from - c_rec
             ps_stacker.add(c_rec)
@@ -1405,7 +1468,7 @@ class MLGPRForegroundFitter(fgfit.AbstractForegroundFitter):
             self.config.save(os.path.join(save_dir, save_name + '.config.parset'))
 
         sampler.run_mcmc(self.config.mcmc.n_steps, self.config.mcmc.n_walkers, verbose=verbose,
-                         live_update=live_update, save_file=save_file)
+                         live_update=live_update, save_file=save_file, emcee_moves=self.config.mcmc.move)
 
         return MLGPRResult(sampler, self.config)
 
