@@ -297,7 +297,7 @@ class PowerSpectraBuilder(object):
         elif not isinstance(ps_config, PowerSpectraConfig):
             ps_config = PowerSpectraConfig.load(ps_config)
 
-        if not isinstance(eor_bin_list, EorBinList):
+        if eor_bin_list is not None and not isinstance(eor_bin_list, EorBinList):
             eor_bin_list = EorBinList.load(eor_bin_list)
 
         self.ps_config = ps_config
@@ -2122,9 +2122,12 @@ class PsStacker(object):
         self.get_variance().save(os.path.join(dir_path, f'{name}.variance.h5'))
         self.get_ps().save(os.path.join(dir_path, f'{name}.ps.h5'))
 
-    def load(self, dir_path, name):
+    @staticmethod
+    def load(dir_path, name):
         ps_stacker = PsStacker(None, None)
         ps_stacker.all_ps3d = SphericalPowerSpectraMC.load(os.path.join(dir_path, f'{name}.ps3d.h5')).all_ps
         ps_stacker.all_ps2d = CylindricalPowerSpectraMC.load(os.path.join(dir_path, f'{name}.ps2d.h5')).all_ps
         ps_stacker.all_ps = SpatialPowerSpectraMC.load(os.path.join(dir_path, f'{name}.ps.h5')).all_ps
         ps_stacker.all_var = VarianceMC.load(os.path.join(dir_path, f'{name}.variance.h5')).all_ps
+
+        return ps_stacker
