@@ -1,6 +1,7 @@
 #!/usr/bin/env python
 
 import os
+import sys
 import time
 import itertools
 
@@ -1032,30 +1033,33 @@ class FloatRangeType(click.ParamType):
 
 
 @main.command('simu_uv')
-@click.argument('instru', type=click.Choice(['ska_low', 'nenufar', 'lofar_hba'], case_sensitive=False))
+@click.argument('instru', type=click.Choice(['ska_low', 'nenufar', 'lofar_hba', 'nenufar_full', 'a12_hba'], case_sensitive=False))
 @click.argument('z', type=FloatRangeType())
 @click.option('--out_filename', '-o', help='Base filename of the output file', type=str, default='simu_uv',
               show_default=True)
 @click.option('--umin', help='Minimum baseline', default=None, type=int)
 @click.option('--umax', help='Maximum baseline', default=None, type=int)
 @click.option('--dec_deg', help='Target declination, in degrees', default=90, type=int, show_default=True)
-@click.option('--fov', help='Field of View side size (in deg)', default=None, type=float)
+@click.option('--fov', help='Field of View side size (in deg), or "pb_fwhm" to set the FoV as the FWHM of the primary beam.',
+              default=None, type=str)
 @click.option('--bandwidth', help='Bandwidth, in MHz', default=51 * 195.3 * 1e-3, type=float, show_default=True)
 @click.option('--df', help='Spectral resolution, in kHz', default=195.3, type=float, show_default=True)
 @click.option('--total_time', help='Total observing time (in hours)', default=10, type=float, show_default=True)
 @click.option('--int_time', help='Integration time (in second)', default=100, type=float, show_default=True)
 @click.option('--res', help='Image resolution in degrees. If not given a default 4x oversampling will be used',
               default=None, type=float)
+@click.option('--allow_intra_baselines', help='Allow baselines which share the same electronic cabinet.', is_flag=True)
 # z float or min-max-step
 # freqs range between fmhz_min - bw/2 and fmhz_max +- bw/2
 # output on file for each redshift
-def simu_uv(instru, z, out_filename, umin, umax, dec_deg, fov, bandwidth, df, total_time, int_time, res):
+def simu_uv(instru, z, out_filename, umin, umax, dec_deg, fov, bandwidth, df, total_time, int_time, 
+            res, allow_intra_baselines):
     ''' Compute gridded UV coverage at redshift Z for instrument INSTRU.
 
     INSTRU can be any of: ska_low, nenufar or lofar_hba.
     Z can be a single number or two numbers separated by ','
     '''
-    from ps_eor import psutil, obssimu
+    from ps_eor import psutil, obssimu, datacube
 
     assert total_time < 24, 'total_time of one night must be < 24 hours'
 
@@ -1069,17 +1073,27 @@ def simu_uv(instru, z, out_filename, umin, umax, dec_deg, fov, bandwidth, df, to
         freqs = np.arange(fmhz - bandwidth / 2, fmhz + bandwidth / 2, df * 1e-3) * 1e6
         z_str = f'{z:.1f}'
 
-    telescop = obssimu.TelescopeSimu.from_name(instru, freqs, dec_deg, 0, total_time, umin, umax, int_time)
+    telescop = obssimu.TelescopeSimu.from_name(instru, freqs, dec_deg, 0, total_time, umin,
+                                               umax, int_time, not allow_intra_baselines)
 
     if fov is None:
         fov = telescop.fov
+    elif fov == 'pb_fwhm':
+        pb = datacube.PrimaryBeam.from_name(telescop.pb_name)
+        fov = np.degrees(pb.get_fwhm(freqs.mean()))
+    else:
+        try:
+            fov = float(fov)
+        except Exception:
+            print('fov should be a float or "pb_fwhm')
+            return 1
 
     print(f'Computing gridded UV configuration for instrument {telescop.name}')
     print(f'Simulating {total_time} hour long observation with {int_time} s integration time')
     print(f'Redshift z = {z_str}')
     print(f'{len(freqs)} frequency channels between {freqs.min() * 1e-6:.1f} and {freqs.max() * 1e-6:.1f} MHz')
     print(f'Frequency resolution: {df:.1f} kHz')
-    print(f'Baseline range: {telescop.umin} - {telescop.umax} lambda')
+    print(f'Baseline range: {telescop.umin} - {telescop.umax} lambda (discarding intra-baselines: {telescop.remove_intra_baselines})')
     print(f'FoV: {fov:.1f} degrees')
 
     uu_meter, vv_meter, _ = telescop.simu_uv()
@@ -1156,6 +1170,10 @@ def simu_noise_ps(uv_simu_h5, total_time_hour, out_filename, kmin, kmax, nks,
     uv_simu = obssimu.TelescopGriddedUV.load(uv_simu_h5)
     freqs = uv_simu.weights.freqs
     ps_gen = uv_simu.get_ps_gen(filter_kpar_min, filter_wedge_theta)
+    telescop = uv_simu.telescope_simu
+
+    print(f'Using telescope configuration {telescop.name}.')
+    print(f'Baseline range: {telescop.umin} - {telescop.umax} lambda (discarding intra-baselines: {telescop.remove_intra_baselines})')
 
     if kmin is None:
         kmin = ps_gen.kmin
@@ -1248,7 +1266,6 @@ def simu_noise_ps_zrange(uv_simu_h5, total_time_hour, k, out_filename, bandwidth
 
     if smooth:
         w_l = min(int(psutil.get_next_odd(2 / delta_z)), int(psutil.get_previous_odd(len(zs))))
-        print(w_l)
         pks = 10 ** savgol_filter(np.log10(pks), w_l, 1)
         pks_err = 10 ** savgol_filter(np.log10(pks_err), w_l, 1)
 
